import json
import os
import numpy as np

import rlcard
from rlcard.envs.env import Env
from rlcard.games.leducholdem.game import LeducholdemGame as Game
from rlcard.utils.utils import *
from rlcard import models


class LeducholdemEnv(Env):
    ''' Limitholdem Environment
    '''

    def __init__(self, allow_step_back=False):
        ''' Initialize the Limitholdem environment
        '''
        super().__init__(Game(allow_step_back), allow_step_back)
        self.actions = ['call', 'raise', 'fold', 'check']
        self.state_shape = [6]

        with open(os.path.join(rlcard.__path__[0], 'games/leducholdem/card2index.json'), 'r') as file:
            self.card2index = json.load(file)

    def print_state(self, player):
        ''' Print out the state of a given player

        Args:
            player (int): Player id
        '''
        state = self.game.get_state(player)
        print('\n=============== Community Card ===============')
        print_card(state['public_card'])
        print('===============   Your Hand    ===============')
        print_card(state['hand'])
        print('===============     Chips      ===============')
        print('Yours:   ', end='')
        for _ in range(state['my_chips']):
            print('+', end='')
        print('')
        for i in range(self.player_num):
            if i != self.active_player:
                print('Agent {}: '.format(i) , end='')
                for _ in range(state['all_chips'][i]):
                    print('+', end='')
        print('\n=========== Actions You Can Choose ===========')
        print(', '.join([str(self.actions.index(action)) + ': ' + action for action in state['legal_actions']]))
        print('')

    def print_result(self, player):
        ''' Print the game result when the game is over

        Args:
            player (int): The human player id
        '''
        payoffs = self.get_payoffs()
        hands = [self.game.players[i].hand.get_index() for i in range(self.player_num)]
        print('')

        if not self.game.players[self.active_player].status == 'fold':
            for i in range(self.player_num):
                if i != self.active_player:
                    print('===============     Agent {}    ==============='.format(i))
                    print_card(hands[i])

        print('===============     Result     ===============')
        if payoffs[player] > 0:
            print('You win {} chips!'.format(payoffs[player]))
        elif payoffs[player] == 0:
            print('It is a tie.')
        else:
            print('You lose {} chips!'.format(-payoffs[player]))
        print('')

    @staticmethod
    def print_action(action):
        ''' Print out an action in a nice form

        Args:
            action (str): A string a action
        '''
        print(action, end='')

    def load_model(self):
        ''' Load pretrained/rule model

        Returns:
            model (Model): A Model object
        '''
        return models.load('leduc-holdem-nfsp')

    def get_legal_actions(self):
        ''' Get all leagal actions

        Returns:
            encoded_action_list (list): return encoded legal action list (from str to int)
        '''
        return self.game.get_legal_actions()

    def extract_state(self, state):
        ''' Extract the state representation from state dictionary for agent

        Note: Currently the use the hand cards and the public cards. TODO: encode the states

        Args:
            state (dict): Original state from the game

        Returns:
            observation (list): combine the player's score and dealer's observable score for observation
        '''
        processed_state = {}

        legal_actions = [self.actions.index(a) for a in state['legal_actions']]
        processed_state['legal_actions'] = legal_actions

        public_card = state['public_card']
        hand = state['hand']
        cards = [] + [hand]
        if public_card:
            cards.append(public_card)
        idx = [self.card2index[card] for card in cards]
        obs = np.zeros(6)
        obs[idx] = 1
        processed_state['obs'] = obs

        return processed_state

    def get_payoffs(self):
        ''' Get the payoff of a game

        Returns:
           payoffs (list): list of payoffs
        '''
        return self.game.get_payoffs()

    def decode_action(self, action_id):
        ''' Decode the action for applying to the game

        Args:
            action id (int): action id

        Returns:
            action (str): action for the game
        '''
        legal_actions = self.game.get_legal_actions()
        if self.actions[action_id] not in legal_actions:
            if 'check' in legal_actions:
                return 'check'
            else:
                return 'fold'
        return self.actions[action_id]



import numpy as np
import random

from rlcard.utils.utils import *

class Env(object):
    ''' The base Env class
    '''

    def __init__(self, game, allow_step_back=False):
        ''' Initialize

        Args:
            game (Game): The Game class
            allow_step_back (boolean): True if allowing step_back
        '''
        self.name = None
        self.game = game
        self.allow_step_back = allow_step_back

        # Get number of players/actions in this game
        self.player_num = game.get_player_num()
        self.action_num = game.get_action_num()

        # A counter for the timesteps
        self.timestep = 0

        # MODES
        self.single_agent_mode = False
        self.active_player = None
        self.human_mode = False


    def init_game(self):
        ''' Start a new game

        Returns:
            (tuple): Tuple containing:

                (numpy.array): The begining state of the game
                (int): The begining player
        '''
        state, player_id = self.game.init_game()
        return self.extract_state(state), player_id

    def step(self, action):
        ''' Step forward

        Args:
            action (int): the action taken by the current player

        Returns:
            (tuple): Tuple containing:

                (numpy.array): The next state
                (int): The ID of the next player
        '''
        if self.single_agent_mode or self.human_mode:
            return self.single_agent_step(action)

        self.timestep += 1
        next_state, player_id = self.game.step(self.decode_action(action))

        return self.extract_state(next_state), player_id

    def single_agent_step(self, action):
        ''' Step forward for human/single agent

        Args:
            action (int): the action takem by the current player

        Returns:
            next_state (numpy.array): The next state
        '''
        reward = 0.
        done = False
        self.timestep += 1
        state, player_id = self.game.step(self.decode_action(action))
        while not self.game.is_over() and not player_id == self.active_player:
            self.timestep += 1
            if self.model.use_raw:
                action = self.model.agents[player_id].eval_step(state)
            else:
                action = self.model.agents[player_id].eval_step(self.extract_state(state))
                action = self.decode_action(action)
            if self.human_mode:
                print('\r>> Agent {} chooses '.format(player_id), end='')
                self.print_action(action)
                print('')
            state, player_id = self.game.step(action)

        if self.game.is_over():
            reward = self.get_payoffs()[self.active_player]
            done = True
            if self.human_mode:
                self.print_result(self.active_player)
            self.reset()

        elif self.human_mode:
            self.print_state(self.active_player)

        return self.extract_state(state), reward, done

    def reset(self):
        ''' Reset environment in single-agent mode
        '''
        if not self.single_agent_mode and not self.human_mode:
            raise ValueError('Reset can only be used in single-agent mode or human mode')

        if self.human_mode:
            history = []
        while True:
            state, player_id = self.game.init_game()
            while not player_id == self.active_player:
                self.timestep += 1
                if self.model.use_raw:
                    action = self.model.agents[player_id].eval_step(state)
                else:
                    action = self.model.agents[player_id].eval_step(self.extract_state(state))
                    action = self.decode_action(action)
                if self.human_mode:
                    history.append((player_id, action))
                state, player_id = self.game.step(action)

            if not self.game.is_over():
                if self.human_mode:
                    print('\n>> Start a new game!')
                    for player_id, action in history:
                        print('\r>> Agent {} chooses '.format(player_id), end='')
                        self.print_action(action)
                        print('')
                    self.print_state(self.active_player)
                break
            else:
                if self.human_mode:
                    history.clear()

        return self.extract_state(state)

    def step_back(self):
        ''' Take one step backward.

        Returns:
            (tuple): Tuple containing:

                (numpy.array): The previous state
                (int): The ID of the previous player

        Note: Error will be raised if step back from the root node.
        '''
        if not self.allow_step_back:
            raise Exception('Step back is off. To use step_back, please set allow_step_back=True in rlcard.make')

        if not self.game.step_back():
            return False

        player_id = self.get_player_id()
        state = self.get_state(player_id)

        return state, player_id


    def get_player_id(self):
        ''' Get the current player id

        Returns:
            (int): the id of the current player
        '''
        return self.game.get_player_id()

    def is_over(self):
        ''' Check whether the curent game is over

        Returns:
            (boolean): True is current game is over
        '''
        return self.game.is_over()

    def get_state(self, player_id):
        ''' Get the state given player id

        Args:
            player_id (int): The player id

        Returns:
            (numpy.array): The observed state of the player
        '''
        return self.extract_state(self.game.get_state(player_id))

    def set_agents(self, agents):
        ''' Set the agents that will interact with the environment

        Args:
            agents (list): List of Agent classes
        '''
        if self.single_agent_mode or self.human_mode:
            raise ValueError('Setting agent in single agent mode or human mode is not allowed.')

        self.agents = agents

    def run(self, is_training=False, seed=None):
        ''' Run a complete game, either for evaluation or training RL agent.

        Args:
            is_training (boolean): True if for training purpose.
            seed (int): A seed for running the game. For single-process program,
              the seed should be set to None. For multi-process program, the
              seed should be asigned for reproducibility.

        Returns:
            (tuple) Tuple containing:

                (list): A list of trajectories generated from the environment.
                (list): A list payoffs. Each entry corresponds to one player.

        Note: The trajectories are 3-dimension list. The first dimension is for different players.
              The second dimension is for different transitions. The third dimension is for the contents of each transiton
        '''
        if self.single_agent_mode or self.human_mode:
            raise ValueError('Run in single agent mode or human mode is not allowed.')

        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)

        trajectories = [[] for _ in range(self.player_num)]
        state, player_id = self.init_game()

        # Loop to play the game
        trajectories[player_id].append(state)
        while not self.is_over():
            # Agent plays
            if not is_training:
                action = self.agents[player_id].eval_step(state)
            else:
                action = self.agents[player_id].step(state)

            # Environment steps
            next_state, next_player_id = self.step(action)
            # Save action
            trajectories[player_id].append(action)

            # Set the state and player
            state = next_state
            player_id = next_player_id

            # Save state.
            if not self.game.is_over():
                trajectories[player_id].append(state)

        # Add a final state to all the players
        for player_id in range(self.player_num):
            state = self.get_state(player_id)
            trajectories[player_id].append(state)

        # Payoffs
        payoffs = self.get_payoffs()

        # Reorganize the trajectories
        trajectories = reorganize(trajectories, payoffs)

        return trajectories, payoffs

    def run_multi(self, task_num, result, is_training=False, seed=None):
        if seed is not None:
            np.random.seed(seed)
        for _ in range(task_num):
            result.append(self.run(is_training=is_training))

    def set_mode(self, active_player=0, single_agent_mode=False, human_mode=False):
        ''' Turn on the single-agent-mode. Pretrained models will
            be loaded to simulate other agents

        Args:
            active_player (int): The player that does not use pretrained models
        '''
        if not isinstance(active_player, int) or active_player < 0 or active_player >= self.player_num:
            raise ValueError('Active player should be a positiv integer less than the player number')

        if not single_agent_mode and not human_mode:
            raise ValueError('You must set single_agent_mode=True, or human_mode=True')

        if single_agent_mode and human_mode:
            raise ValueError('You can not set single_agentmode=True and human_mode=True together/')

        self.model = self.load_model()
        self.active_player = active_player
        self.single_agent_mode = single_agent_mode
        self.human_mode = human_mode

    def print_state(self, player):
        ''' Print out the state of a given player

        Args:
            player (int): Player id
        '''
        raise NotImplementedError

    def print_result(self, player):
        ''' Print the game result when the game is over

        Args:
            player (int): The human player id
        '''
        raise NotImplementedError

    @staticmethod
    def print_action(action):
        ''' Print out an action in a nice form

        Args:
            action (str): A string a action
        '''
        raise NotImplementedError

    def load_model(self):
        ''' Load pretrained/rule model

        Returns:
            model (Model): A Model object
        '''
        raise NotImplementedError

    def extract_state(self, state):
        ''' Extract useful information from state for RL. Must be implemented in the child class.

        Args:
            state (dict): the raw state

        Returns:
            (numpy.array): the extracted state
        '''
        raise NotImplementedError

    def get_payoffs(self):
        ''' Get the payoffs of players. Must be implemented in the child class.

        Returns:
            (list): A list of payoffs for each player.

        Note: Must be implemented in the child class.
        '''
        raise NotImplementedError

    def decode_action(self, action_id):
        ''' Decode Action id to the action in the game.

        Args:
            action_id (int): The id of the action

        Returns:
            (string): The action that will be passed to the game engine.

        Note: Must be implemented in the child class.
        '''
        raise NotImplementedError

    def get_legal_actions(self):
        ''' Get all legal actions for current state.

        Returns:
            (list): A list of legal actions' id.

        Note: Must be implemented in the child class.
        '''
        raise NotImplementedError



''' Register new environments
'''

from rlcard.envs.registration import register, make

register(
    env_id='blackjack',
    entry_point='rlcard.envs.blackjack:BlackjackEnv',
)

register(
    env_id='doudizhu',
    entry_point='rlcard.envs.doudizhu:DoudizhuEnv',
)

register(
    env_id='limit-holdem',
    entry_point='rlcard.envs.limitholdem:LimitholdemEnv',
)

register(
    env_id='no-limit-holdem',
    entry_point='rlcard.envs.nolimitholdem:NolimitholdemEnv',
)

register(
    env_id='leduc-holdem',
    entry_point='rlcard.envs.leducholdem:LeducholdemEnv'
)

register(
    env_id='uno',
    entry_point='rlcard.envs.uno:UnoEnv',
)

register(
    env_id='mahjong',
    entry_point='rlcard.envs.mahjong:MahjongEnv',
)



import numpy as np

from rlcard.utils.utils import *
from rlcard.envs.env import Env
from rlcard.games.mahjong.game import MahjongGame as Game
from rlcard.games.mahjong.card import MahjongCard as Card
from rlcard.games.mahjong.utils import *

class MahjongEnv(Env):
    ''' Mahjong Environment
    '''

    def __init__(self, allow_step_back=False):
        super().__init__(Game(allow_step_back), allow_step_back)
        self.action_id = card_encoding_dict
        self.de_action_id = {self.action_id[key]: key for key in self.action_id.keys()}
        self.state_shape = [6, 34, 4]

    def extract_state(self, state):
        ''' Encode state

        Args:
            state (dict): dict of original state

        Returns:
            numpy array: 6*5*15 array
                         6 : current hand
                             the union of the other two players' hand
                             the recent three actions
                             the union of all played cards
        '''
        players_pile = state['players_pile']
        hand_rep = encode_cards(state['current_hand'])
        piles_rep = []
        for p in players_pile.keys():
            piles_rep.append(encode_cards(pile2list(players_pile[p])))
        piles_rep = np.array(piles_rep)
        table_rep = encode_cards(state['table'])
        rep = [hand_rep, table_rep]
        rep.extend(piles_rep)
        obs = np.array(rep)

        extrated_state = {'obs': obs, 'legal_actions': self.get_legal_actions()}
        return extrated_state

    def get_payoffs(self):
        ''' Get the payoffs of players. Must be implemented in the child class.

        Returns:
            payoffs (list): a list of payoffs for each player
        '''
        _, player, _ = self.game.judger.judge_game(self.game)
        if player == -1:
            payoffs = [0, 0, 0, 0]
        else:
            payoffs = [-1, -1, -1, -1]
            payoffs[player] = 1
        return payoffs

    def decode_action(self, action_id):
        ''' Action id -> the action in the game. Must be implemented in the child class.

        Args:
            action_id (int): the id of the action

        Returns:
            action (string): the action that will be passed to the game engine.
        '''
        action = self.de_action_id[action_id]
        if action_id < 34:
            candidates = self.game.get_legal_actions(self.game.get_state(self.game.round.current_player))
            for card in candidates:
                if card.get_str() == action:
                    action = card
                    break
        return action

    def get_legal_actions(self):
        ''' Get all legal actions for current state

        Returns:
        if type(legal_actions[0]) == Card:
            print("GET:", [c.get_str() for c in legal_actions])
        else:
            print(legal_actions)
            legal_actions (list): a list of legal actions' id
        '''
        legal_action_id = []
        legal_actions = self.game.get_legal_actions(self.game.get_state(self.game.round.current_player))
        if legal_actions:
            for action in legal_actions:
                if isinstance(action, Card):
                    action = action.get_str()
                action_id = self.action_id[action]
                legal_action_id.append(action_id)
        else:
            print("##########################")
            print("No Legal Actions")
            print(self.game.judger.judge_game(self.game))
            print(self.game.is_over())
            print([len(p.pile) for p in self.game.players])
            #print(self.game.get_state(self.game.round.current_player))
            #exit()
        return legal_action_id



import numpy as np

from rlcard.envs.env import Env
from rlcard import models
from rlcard.games.uno.game import UnoGame as Game
from rlcard.games.uno.utils import encode_hand, encode_target
from rlcard.games.uno.utils import ACTION_SPACE, ACTION_LIST
from rlcard.games.uno.card import UnoCard


class UnoEnv(Env):

    def __init__(self, allow_step_back=False):
        super().__init__(Game(allow_step_back), allow_step_back)
        self.state_shape = [7, 4, 15]

    def print_state(self, player):
        ''' Print out the state of a given player

        Args:
            player (int): Player id
        '''
        state = self.game.get_state(player)
        print('\n=============== Your Hand ===============')
        UnoCard.print_cards(state['hand'])
        print('')
        print('=============== Last Card ===============')
        UnoCard.print_cards(state['target'], wild_color=True)
        print('')
        print('========== Agents Card Number ===========')
        for i in range(self.player_num):
            if i != self.active_player:
                print('Agent {} has {} cards.'.format(i, len(self.game.players[i].hand)))
        print('======== Actions You Can Choose =========')
        for i, action in enumerate(state['legal_actions']):
            print(str(ACTION_SPACE[action])+': ', end='')
            UnoCard.print_cards(action, wild_color=True)
            if i < len(state['legal_actions']) - 1:
                print(', ', end='')
        print('\n')

    def print_result(self, player):
        ''' Print the game result when the game is over

        Args:
            player (int): The human player id
        '''
        payoffs = self.get_payoffs()
        print('===============     Result     ===============')
        if payoffs[player] > 0:
            print('You win!')
        else:
            print('You lose!')
        print('')

    @staticmethod
    def print_action(action):
        ''' Print out an action in a nice form

        Args:
            action (str): A string a action
        '''
        UnoCard.print_cards(action, wild_color=True)

    def load_model(self):
        ''' Load pretrained/rule model

        Returns:
            model (Model): A Model object
        '''
        return models.load('uno-rule-v1')

    def extract_state(self, state):
        obs = np.zeros((7, 4, 15), dtype=int)
        encode_hand(obs[:3], state['hand'])
        encode_target(obs[3], state['target'])
        encode_hand(obs[4:], state['others_hand'])
        legal_action_id = self.get_legal_actions()
        extrated_state = {'obs': obs, 'legal_actions': legal_action_id}
        return extrated_state

    def get_payoffs(self):

        return self.game.get_payoffs()

    def decode_action(self, action_id):
        legal_ids = self.get_legal_actions()
        if action_id in legal_ids:
            return ACTION_LIST[action_id]
        #if (len(self.game.dealer.deck) + len(self.game.round.played_cards)) > 17:
        #    return ACTION_LIST[60]
        return ACTION_LIST[np.random.choice(legal_ids)]

    def get_legal_actions(self):
        legal_actions = self.game.get_legal_actions()
        legal_ids = [ACTION_SPACE[action] for action in legal_actions]
        return legal_ids



import importlib

class EnvSpec(object):
    ''' A specification for a particular instance of the environment.
    '''

    def __init__(self, env_id, entry_point=None):
        ''' Initilize

        Args:
            env_id (string): the name of the environent
            entry_point (string): a string the indicates the location of the envronment class
        '''
        self.env_id = env_id
        mod_name, class_name = entry_point.split(':')
        self._entry_point = getattr(importlib.import_module(mod_name), class_name)

    def make(self, allow_step_back=False):
        ''' Instantiates an instance of the environment

        Returns:
            env (Env): an instance of the environemnt
            allow_step_back (boolean): True if you wants to able to step_back
        '''
        env = self._entry_point(allow_step_back)
        return env


class EnvRegistry(object):
    ''' Register an environment (game) by ID
    '''

    def __init__(self):
        ''' Initilize
        '''
        self.env_specs = {}

    def register(self, env_id, entry_point):
        ''' Register an environment

        Args:
            env_id (string): the name of the environent
            entry_point (string): a string the indicates the location of the envronment class
        '''
        if env_id in self.env_specs:
            raise ValueError('Cannot re-register env_id: {}'.format(env_id))
        self.env_specs[env_id] = EnvSpec(env_id, entry_point)

    def make(self, env_id, allow_step_back=False):
        ''' Create and environment instance

        Args:
            env_id (string): the name of the environment
            allow_step_back (boolean): True if you wants to able to step_back
        '''
        if env_id not in self.env_specs:
            raise ValueError('Cannot find env_id: {}'.format(env_id))
        return self.env_specs[env_id].make(allow_step_back)

# Have a global registry
registry = EnvRegistry()


def register(env_id, entry_point):
    ''' Register an environment

    Args:
        env_id (string): the name of the environent
        entry_point (string): a string the indicates the location of the envronment class
    '''
    return registry.register(env_id, entry_point)

def make(env_id, allow_step_back=False):
    ''' Create and environment instance

    Args:
        env_id (string): the name of the environment
        allow_step_back (boolean): True if you wants to able to step_back
    '''
    return registry.make(env_id, allow_step_back)



import matplotlib.pyplot as plt
import os

class Logger(object):
    ''' Logger saves the running results and helps make plots from the results
    '''

    def __init__(self, xlabel = '', ylabel = '', legend = '', log_path = None, csv_path = None):
        ''' Initialize the labels, legend and paths of the plot and log file.

        Args:
            xlabel (string): label of x axis of the plot
            ylabel (string): label of y axis of the plot
            legend (string): name of the curve
            log_path (string): where to store the log file
            csv_path (string): where to store the csv file

        Note:
            1. log_path must be provided to use the log() method. If the log file already exists, it will be deleted when Logger is initialized.
            2. If csv_path is provided, then one record will be write to the file everytime add_point() method is called.
        '''
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.legend = legend
        self.xs = []
        self.ys = []
        self.log_path = log_path
        self.csv_path = csv_path
        self.log_file = None
        self.csv_file = None
        if log_path != None:
            log_dir = os.path.dirname(log_path)
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)
            self.log_file = open(log_path, 'w')
        if csv_path != None:
            csv_dir = os.path.dirname(csv_path)
            if not os.path.exists(csv_dir):
                os.makedirs(csv_dir)
            self.csv_file = open(csv_path, 'w')
            self.csv_file.write(xlabel+','+ylabel+'\n')
            self.csv_file.flush()

    def log(self, text):
        ''' Write the text to log file then print it.

        Args:
            text(string): text to log
        '''
        self.log_file.write(text+'\n')
        self.log_file.flush()
        print(text)

    def add_point(self, x = None, y = None):
        ''' Add a point to the plot

        Args:
            x (Number): x coordinate value
            y (Number): y coordinate value
        '''
        if x != None and y != None:
            self.xs.append(x)
            self.ys.append(y)
        else:
            raise ValueError('x and y should not be None.')

        # If csv_path is not None then write x and y to file
        if self.csv_path != None:
            self.csv_file.write(str(x)+','+str(y)+'\n')
            self.csv_file.flush()

    def make_plot(self, save_path = ''):
        ''' Make plot using all stored points

        Args:
            save_path (string): where to store the plot
        '''
        fig, ax = plt.subplots()
        ax.plot(self.xs, self.ys, label=self.legend)
        ax.set(xlabel=self.xlabel, ylabel=self.ylabel)
        ax.legend()
        ax.grid()

        save_dir = os.path.dirname(save_path)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        fig.savefig(save_path)

    def close_file(self):
        ''' Close the created file objects
        '''
        if self.log_path != None:
            self.log_file.close()
        if self.csv_path != None:
            self.csv_file.close()



import numpy as np

from rlcard.utils.utils import *
from rlcard.envs.env import Env
from rlcard.games.doudizhu.game import DoudizhuGame as Game
from rlcard.games.doudizhu.utils import SPECIFIC_MAP
from rlcard.games.doudizhu.utils import ACTION_LIST, ACTION_SPACE
from rlcard.games.doudizhu.utils import encode_cards


class DoudizhuEnv(Env):
    ''' Doudizhu Environment
    '''

    def __init__(self, allow_step_back=False):
        super().__init__(Game(allow_step_back), allow_step_back)
        self.state_shape = [6, 5, 15]

    def extract_state(self, state):
        ''' Encode state

        Args:
            state (dict): dict of original state

        Returns:
            numpy array: 6*5*15 array
                         6 : current hand
                             the union of the other two players' hand
                             the recent three actions
                             the union of all played cards
        '''
        obs = np.zeros((6, 5, 15), dtype=int)
        for index in range(6):
            obs[index][0] = np.ones(15, dtype=int)
        encode_cards(obs[0], state['current_hand'])
        encode_cards(obs[1], state['others_hand'])
        for i, action in enumerate(state['trace'][-3:]):
            if action[1] != 'pass':
                encode_cards(obs[4-i], action[1])
        if state['played_cards'] is not None:
            encode_cards(obs[5], state['played_cards'])

        extrated_state = {'obs': obs, 'legal_actions': self.get_legal_actions()}
        return extrated_state

    def get_payoffs(self):
        ''' Get the payoffs of players. Must be implemented in the child class.

        Returns:
            payoffs (list): a list of payoffs for each player
        '''
        return self.game.judger.judge_payoffs(self.game.round.landlord_id, self.game.winner_id)

    def decode_action(self, action_id):
        ''' Action id -> the action in the game. Must be implemented in the child class.

        Args:
            action_id (int): the id of the action

        Returns:
            action (string): the action that will be passed to the game engine.
        '''
        abstract_action = ACTION_LIST[action_id]
        legal_actions = self.game.state['actions']
        specific_actions = []
        for legal_action in legal_actions:
            for abstract in SPECIFIC_MAP[legal_action]:
                if abstract == abstract_action:
                    specific_actions.append(legal_action)
        if specific_actions:
            action = np.random.choice(specific_actions)
        else:
            if "pass" in legal_actions:
                action = "pass"
            else:
                action = np.random.choice(legal_actions)

        return action

    def get_legal_actions(self):
        ''' Get all legal actions for current state

        Returns:
            legal_actions (list): a list of legal actions' id
        '''
        legal_action_id = []
        legal_actions = self.game.state['actions']
        if legal_actions:
            for action in legal_actions:
                for abstract in SPECIFIC_MAP[action]:
                    action_id = ACTION_SPACE[abstract]
                    if action_id not in legal_action_id:
                        legal_action_id.append(action_id)
        return legal_action_id



import json
import os
import numpy as np

import rlcard
from rlcard.envs.env import Env
from rlcard.games.nolimitholdem.game import NolimitholdemGame as Game
from rlcard.utils.utils import *

class NolimitholdemEnv(Env):
    ''' Limitholdem Environment
    '''

    def __init__(self, allow_step_back=False):
        ''' Initialize the Limitholdem environment
        '''
        super().__init__(Game(allow_step_back), allow_step_back)
        self.actions = ['call', 'fold', 'check']
        self.state_shape = [54]
        for raise_amount in range(1, self.game.init_chips+1):
            self.actions.append(raise_amount)

        with open(os.path.join(rlcard.__path__[0], 'games/limitholdem/card2index.json'), 'r') as file:
            self.card2index = json.load(file)

    def get_legal_actions(self):
        ''' Get all leagal actions

        Returns:
            encoded_action_list (list): return encoded legal action list (from str to int)
        '''
        return self.game.get_legal_actions()

    def extract_state(self, state):
        ''' Extract the state representation from state dictionary for agent

        Note: Currently the use the hand cards and the public cards. TODO: encode the states

        Args:
            state (dict): Original state from the game

        Returns:
            observation (list): combine the player's score and dealer's observable score for observation
        '''
        processed_state = {}

        legal_actions = [self.actions.index(a) for a in state['legal_actions']]
        processed_state['legal_actions'] = legal_actions

        public_cards = state['public_cards']
        hand = state['hand']
        my_chips = state['my_chips']
        all_chips = state['all_chips']
        cards = public_cards + hand
        idx = [self.card2index[card] for card in cards]
        obs = np.zeros(54)
        obs[idx] = 1
        obs[52] = float(my_chips)
        obs[53] = float(max(all_chips))
        processed_state['obs'] = obs

        return processed_state

    def get_payoffs(self):
        ''' Get the payoff of a game

        Returns:
           payoffs (list): list of payoffs
        '''
        return self.game.get_payoffs()

    def decode_action(self, action_id):
        ''' Decode the action for applying to the game

        Args:
            action id (int): action id

        Returns:
            action (str): action for the game
        '''
        legal_actions = self.game.get_legal_actions()
        if self.actions[action_id] not in legal_actions:
            if 'check' in legal_actions:
                return 'check'
            else:
                return 'fold'
        return self.actions[action_id]



import setuptools

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rlcard",
    version="0.1",
    author="Data Analytics at Texas A&M (DATA) Lab",
    author_email="khlai037@tamu.edu",
    description="A Toolkit for Reinforcement Learning in Card Games",
    url="https://github.com/datamllab/rlcard",
    keywords=["Reinforcement Learning", "game", "RL", "AI"],
    packages=setuptools.find_packages(),
    install_requires=[
        'tensorflow>=1.14,<2.0',
        'tensorflow_probability==0.7.0',
        'dm-sonnet==1.35',
        'numpy>=1.16.3',
        'matplotlib>=3.0'
    ],
    requires_python='>=3.5',
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import numpy as np

from rlcard.envs.env import Env
from rlcard.games.blackjack.game import BlackjackGame as Game


class BlackjackEnv(Env):
    ''' Blackjack Environment
    '''

    def __init__(self, allow_step_back=False):
        ''' Initialize the Blackjack environment
        '''
        super().__init__(Game(allow_step_back), allow_step_back)
        self.rank2score = {"A":10, "2":2, "3":3, "4":4, "5":5, "6":6, "7":7, "8":8, "9":9, "T":10, "J":10, "Q":10, "K":10}
        self.actions = ['hit', 'stand']
        self.state_shape = [2]

    def get_legal_actions(self):
        ''' Get all leagal actions

        Returns:
            encoded_action_list (list): return encoded legal action list (from str to int)
        '''
        encoded_action_list = []
        for i in range(len(self.actions)):
            encoded_action_list.append(i)
        return encoded_action_list

    def extract_state(self, state):
        ''' Extract the state representation from state dictionary for agent

        Args:
            state (dict): Original state from the game

        Returns:
            observation (list): combine the player's score and dealer's observable score for observation
        '''
        cards = state['state']
        my_cards = cards[0]
        dealer_cards = cards[1]

        def get_scores_and_A(hand):
            score = 0
            has_a = 0
            for card in hand:
                score += self.rank2score[card[1:]]
                if card[1] == 'A':
                    has_a = 1
            if score > 21 and has_a == 1:
                score -= 9
            return score, has_a

        my_score, _ = get_scores_and_A(my_cards)
        dealer_score, _ = get_scores_and_A(dealer_cards)
        obs = np.array([my_score, dealer_score])

        legal_actions = [i for i in range(len(self.actions))]
        extracted_state = {'obs': obs, 'legal_actions': legal_actions}

        return extracted_state

    def get_payoffs(self):
        ''' Get the payoff of a game

        Returns:
           payoffs (list): list of payoffs
        '''
        if self.game.winner['player'] == 0 and self.game.winner['dealer'] == 1:
            return [-1]
        elif self.game.winner['dealer'] == 0 and self.game.winner['player'] == 1:
            return [1]
        elif self.game.winner['player'] == 1 and self.game.winner['dealer'] == 1:
            return [0]

    def decode_action(self, action_id):
        ''' Decode the action for applying to the game

        Args:
            action id (int): action id

        Returns:
            action (str): action for the game
        '''
        return self.actions[action_id]



''' Wrrapers of pretrained models. Designed for Tensorflow.
'''

import os
import tensorflow as tf

import rlcard
from rlcard.agents.nfsp_agent import NFSPAgent
from rlcard.agents.random_agent import RandomAgent
from rlcard.models.model import Model

# Root path of pretrianed models
ROOT_PATH = os.path.join(rlcard.__path__[0], 'models/pretrained')

class LeducHoldemNFSPModel(Model):
    ''' A pretrained model on Leduc Holdem with NFSP
    '''

    def __init__(self):
        ''' Load pretrained model
        '''
        self.graph = tf.Graph()
        self.sess = tf.Session(graph=self.graph)

        env = rlcard.make('leduc-holdem')
        with self.graph.as_default():
            self.nfsp_agents = []
            for i in range(env.player_num):
                agent = NFSPAgent(self.sess,
                                  scope='nfsp' + str(i),
                                  action_num=env.action_num,
                                  state_shape=env.state_shape,
                                  hidden_layers_sizes=[128,128],
                                  q_norm_step=1000,
                                  q_mlp_layers=[128,128])
                self.nfsp_agents.append(agent)
            normalize(env, self.nfsp_agents, 1000)
            self.sess.run(tf.global_variables_initializer())

        check_point_path = os.path.join(ROOT_PATH, 'leduc_holdem_nfsp')
        with self.sess.as_default():
            with self.graph.as_default():
                saver = tf.train.Saver(tf.model_variables())
                saver.restore(self.sess, tf.train.latest_checkpoint(check_point_path))
    @property
    def agents(self):
        ''' Get a list of agents for each position in a the game

        Returns:
            agents (list): A list of agents

        Note: Each agent should be just like RL agent with step and eval_step
              functioning well.
        '''
        return self.nfsp_agents

    @property
    def use_raw(self):
        ''' Indicate whether use raw state and action

        Returns:
            use_raw (boolean): True if using raw state and action
        '''
        return False


def normalize(e, agents, num):
    ''' Feed random data to normalizer

    Args:
        e (Env): AN Env class
        agents (list): A list of Agent object
        num (int): The number of steps to be normalized

    '''
    begin_step = e.timestep
    e.set_agents([RandomAgent(e.action_num) for _ in range(e.player_num)])
    while e.timestep - begin_step < num:
        trajectories, _ = e.run(is_training=False)
        for agent in agents:
            for tra in trajectories:
                for ts in tra:
                    agent.feed(ts)

